#
#  Copyright (c) 2011-2014 Exxeleron GmbH
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

__all__ = ['qconnection', 'qtype', 'qtemporal', 'qcollection', 'MetaData', 'CONVERSION_OPTIONS']


__version__ = '2.0.0'



try:
    from .fastutils import uncompress
except:
    __is_cython_enabled__ = False
else:
    __is_cython_enabled__ = True


from .qmetaconversion import MetaData, CONVERSION_OPTIONS
